/*
LMX2582 PLL v1
YO4HFU - Pastorcici Robert; 03.12.2024 

Free to be modified. Please, keep my callsign inside sketch.
*/

// Define LMX2582 register for each BCD switch position (R0 register required for correct operation)
//[3] to be modified if more registers are required to be sent during frequency change

const uint32_t LMX_Reg1[3] = {0x240C21, 0x23089B, 0x00229C};   //R36, R35, R0 - 310.7MHz
const uint32_t LMX_Reg2[3] = {0x240C21, 0x23109F, 0x00229C};   //R36, R35, R0 - 310.7MHz divided by 2
const uint32_t LMX_Reg3[3] = {0x240C41, 0x23099F, 0x00229C};  //R36, R35, R0  - 310.7MHz divided by 3
const uint32_t LMX_Reg4[3] = {0x240C42, 0x23099B, 0x00229C};  //R36, R35, R0  - 310.7MHz divided by 4

// Define LMX2582 initialisation register, all 46 registers calculated by "Tics Pro" software
const uint32_t LMX_INIT_Reg[46] = {
0x460000, //R70
0x450000,
0x440089,
0x400377,
0x3E0000,
0x3D0001,
0x3B0000,
0x3003FC,
0x2F00CF,
0x2E0FA1,
0x2D006B,
0x2C0000,
0x2B0000,
0x2A0000,
0x2900C8,
0x280000, //R40

0x278104, //R39
0x26001E,
0x254000,
0x240C21, //R36
0x23089B, //R35
0x22C3EA,
0x212A0A,
0x20210A,
0x1F0601,
0x1E0034,
0x1D0084,
0x1C2924,
0x190000,
0x180509,
0x178842,
0x162300, //R21

0x14012C, //R20
0x130965,
0x0E0420,
0x0D4000,
0x0C7001,
0x0B0018,
0x0A10D8,
0x090302,
0x081084,
0x0728B2,
0x041943,
0x020500,
0x010808,
0x00229C //RO
}; 
                                                                                                          
// Define ADF4001 register F, R, N values calculated by "ADI PLL Int-N" software
const uint32_t ADF_Reg[3] = {0x1F8092, 0x8, 0x1801}; // +PD, 5mA, PFD=5MHz, 120MHz

// Define BCD switch pinout
const int bcd_pins[4] = {8, 9, 7, 10}; // bcd input "1, 2, 4, 8" 

// Define pins
const int CLK = 0;       // pin 13
const int DATA = 1;      // pin 12
const int ADF_CS = 2;    // pin 11
const int LMX_CS = 3;    // pin 10

// Define variables 
int current_position = -1; 
int last_bcd_states[4] = {HIGH, HIGH, HIGH, HIGH};  //initialize all pins to high

unsigned long lastDebounceTime = 0;  // the last time the output pin was toggled
unsigned long debounceDelay = 50;   // the debounce time; increase if the output flickers

uint32_t current_LMX_Reg[3];  //[3]to be modified if more registers are required to be sent during frequency change


void setup() {
  // Set SPI pins as output
  pinMode(CLK, OUTPUT);
  pinMode(DATA, OUTPUT);
  pinMode(LMX_CS, OUTPUT);
  pinMode(ADF_CS, OUTPUT);
  
  digitalWrite(LMX_CS, HIGH);
  digitalWrite(ADF_CS, HIGH);
  
  // Set BCD pullup
  for (int i = 0; i < 4; i++) {
  pinMode(bcd_pins[i], INPUT_PULLUP);
  }


// Send ADF4001 register values once after power on 24-bit
for (int i = 0; i < 3; i++) {
    uint32_t data = ADF_Reg[i];
    for (int j = 0; j < 24; j++) {
      digitalWrite(CLK, LOW);
      digitalWrite(ADF_CS, LOW);
      digitalWrite(DATA, bitRead(data, 23));
      digitalWrite(CLK, HIGH);
      data <<= 1;
    }

  // Latch ADF_CS high after each 24-bit register is sent
  digitalWrite(DATA, LOW);
  digitalWrite(CLK, LOW);
  delayMicroseconds(10); // add a small delay
  digitalWrite(ADF_CS, HIGH);
  delayMicroseconds(10); // add a small delay
  }
  digitalWrite(ADF_CS, HIGH);

  // Initialize LMX2582
  for (int i = 0; i < 46; i++) {
    uint32_t data = LMX_INIT_Reg[i];
    for (int j = 0; j < 24; j++) {
      digitalWrite(CLK, LOW);
      digitalWrite(LMX_CS, LOW);
      digitalWrite(DATA, bitRead(data, 23));
      digitalWrite(CLK, HIGH);
      data <<= 1;
    }
    
  digitalWrite(DATA, LOW);
  digitalWrite(CLK, LOW);
  delayMicroseconds(10); // add a small delay
  digitalWrite(LMX_CS, HIGH);
  delayMicroseconds(10); // add a small delay
  }
  digitalWrite(LMX_CS, HIGH);
 }


void loop() {
  
  // Check BCD switch position with debounce
  int bcd_states[4] = {digitalRead(bcd_pins[0]), digitalRead(bcd_pins[1]), digitalRead(bcd_pins[2]), digitalRead(bcd_pins[3])};

  // Check if the switch state has changed
  if (bcd_states[0] != last_bcd_states[0] || bcd_states[1] != last_bcd_states[1] || bcd_states[2] != last_bcd_states[2] || bcd_states[3] != last_bcd_states[3]) {
    lastDebounceTime = millis();
  }

  if ((millis() - lastDebounceTime) > debounceDelay) {
    // Debounce complete, read the position of the switch
    int position = bcd_states[0] + bcd_states[1]*2 + bcd_states[2]*4 + bcd_states[3]*8;

    if (position != current_position) {
      
    // Set pins as INPUT or OUTPUT based on BCD position for uWire external access, uC controller "disabled"
    if (position == 15) {         // position 0, BCD complement
        pinMode(CLK, INPUT);
        pinMode(DATA, INPUT);
        pinMode(LMX_CS, INPUT);
        pinMode(ADF_CS, INPUT);
      } else {
        pinMode(CLK, OUTPUT);
        pinMode(DATA, OUTPUT);
        pinMode(LMX_CS, OUTPUT);
        pinMode(ADF_CS, OUTPUT);
      }
      
      // Update current 10 positions BCD switch and last BCD states
      current_position = position;
      
      last_bcd_states[0] = bcd_states[0];
      last_bcd_states[1] = bcd_states[1];
      last_bcd_states[2] = bcd_states[2];
      last_bcd_states[3] = bcd_states[3];
      
        switch (position) {
          
        case 14:
        memcpy(current_LMX_Reg, LMX_Reg1, sizeof(current_LMX_Reg)); //position 1, BCD complement 
        break;
        
        case 13:
        memcpy(current_LMX_Reg, LMX_Reg2, sizeof(current_LMX_Reg)); //position 2, BCD complement
        break;

        case 12:
        memcpy(current_LMX_Reg, LMX_Reg3, sizeof(current_LMX_Reg)); //position 3, BCD complement
        break;

        case 11:
        memcpy(current_LMX_Reg, LMX_Reg4, sizeof(current_LMX_Reg)); //position 4, BCD complement
        break;

       // case 10:
       //  memcpy(current_LMX_Reg, LMX_Reg5, sizeof(current_LMX_Reg)); //position 5, BCD complement
       // break;

       // add more positions if required
    }
    
    // Send LMX2582 registers via SPI
    for (int i = 0; i < 3; i++) {            //"3" to be modified if more registers are required to be sent during frequency change
      uint32_t data = current_LMX_Reg[i];
      for (int j = 0; j < 24; j++) {
        digitalWrite(CLK, LOW);
        digitalWrite(LMX_CS, LOW);
        digitalWrite(DATA, bitRead(data, 23));
        digitalWrite(CLK, HIGH);
        data <<= 1;
   }
      
  // Latch LMX_CS high after each 24-bit register is sent
  digitalWrite(DATA, LOW);
  digitalWrite(CLK, LOW);
  delayMicroseconds(10); // add a small delay
  digitalWrite(LMX_CS, HIGH);
  delayMicroseconds(10); // add a small delay
   } 
  digitalWrite(LMX_CS, HIGH); // register sent, all words sent

  // Update current BCD position
  current_position = position;
  }
  }

  // Save current switch state as last state
  last_bcd_states[0] = bcd_states[0];
  last_bcd_states[1] = bcd_states[1];
  last_bcd_states[2] = bcd_states[2];
  last_bcd_states[3] = bcd_states[3];

  // Delay 
  delay(50);
}
